from twisted.internet.defer import Deferred

from edge import eventlog
from edge_bbox.types import mk_plate_bbox2, QuadI, RectF
from edge.lane_detector import LaneDetector
from edge.landmark import camera

from edge.image.image import ImEngine, ImRgba, ImLoader, OverviewImEngine, ImBuffer
from edge.image.types import ImFooterVars, MemBlob
from edge.service.tz import ITzService
from edge.model.types import Dir, Location
from edge.service.about import IAboutService

from .best import get_tr_item_dist
from .types import DetectorArgs
from ..types import IBaseDetector, DetectorError
from ...model.types import (BaseEvent, Measurement, Waypoint, BestWaypoint,
                            Vehicle, LicensePlate, LicensePlateType, Position)
from ...model import models
from ...services.raas import IRaasService, RaasError
from ...services.trainer import ITrainerService
from ...services.plate_info import IPlateInfoService
from ...services.vs import IVsService, VsImageInfo, VsError
from ...services.clock import IClockService


def _mk_racam_speed(tr_track):
    """
    :type tr_track: TrTrack

    :return: Speed measured by radar and camera.
    :rtype: int | None
    """
    min_tr_item = None
    for index in range(0, len(tr_track), 1):
        tr_item = tr_track[index]  # type: models.TrTrackItem
        if tr_item.ra_item and tr_item.lr_item:
            min_tr_item = tr_item
            break

    max_tr_item = None
    for index in range(len(tr_track)-1, -1, -1):
        tr_item = tr_track[index]  # type: models.TrTrackItem
        if tr_item.ra_item and tr_item.lr_item:
            max_tr_item = tr_item
            break

    if None in [min_tr_item, max_tr_item]:
        return

    elif min_tr_item.ts == max_tr_item.ts:
        return

    min_f = min_tr_item.ra_item.raw_f / 100.0
    min_l = min_tr_item.ra_item.raw_l / 100.0

    max_f = max_tr_item.ra_item.raw_f / 100.0
    max_l = max_tr_item.ra_item.raw_l / 100.0

    dist_f = max_f - min_f
    dist_l = max_l - min_l

    if dist_f >= 0:
        dist_r = pow(pow(dist_f, 2) + pow(dist_l, 2), 0.5)

    else:
        dist_r = - pow(pow(dist_f, 2) + pow(dist_l, 2), 0.5)

    dt_sec = (max_tr_item.ts - min_tr_item.ts) / 1e6
    racam_speed = dist_r / dt_sec * 3.6

    return int(round(racam_speed))


class _TrItemImageInfo(VsImageInfo):
    def __init__(self, tr_item, weight):
        """
        :type tr_item: models.TrTrackItem
        :type weight: int
        """
        VsImageInfo.__init__(self, tr_item.ts)
        self.tr_item = tr_item
        self.weight = weight


class _ImStats:
    def __init__(self):
        self.total = 0
        self.success_arch = 0
        self.success_file = 0
        self.failure = 0


class _Stats:
    def __init__(self):
        self.meas_im = _ImStats()
        self.over_im = _ImStats()

        self.ra_cam_speed_total = 0
        self.ra_absent = 0
        self.ra_cam_speed_absent = 0
        self.ra_cam_speed_invalid_car = 0
        self.ra_cam_speed_invalid_bus = 0


class _BestImProvider:
    def __init__(self, vs_service, meas_footer_h, meas_stats, over_stats, over_enable_arch):
        """
        :type vs_service: IVsService
        :type meas_footer_h: int
        :type meas_stats: _ImStats
        :type over_stats: _ImStats
        :type over_enable_arch: bool
        """
        self._vs = vs_service
        self._im_loader = ImLoader()
        self._im_infos = []

        self._meas_footer_h = meas_footer_h
        self._meas_stats = meas_stats
        self.meas_rgba = ImRgba()

        self._over_enable_arch = over_enable_arch
        self._over_stats = over_stats
        self.over_jpeg = ImBuffer()

    def destroy(self):
        self._im_loader.destroy()

    def setup(self, track):
        """
        :type track: models.Track
        :rtype: None
        """
        def mk_tr_item_image_info(tr_item):
            """
            :type tr_item: models.TrTrackItem
            :rtype: _TrItemImageInfo
            """
            tr_item = models.TrTrackItem(tr_item.ts, tr_item.lr_item, tr_item.ra_item, tr_item.le_item)
            lr_scene_w = lr_scene_h = None
            if track.lr_track:
                lr_scene_w = track.lr_track.best_image.width
                lr_scene_h = track.lr_track.best_image.width

            return _TrItemImageInfo(tr_item, get_tr_item_dist(tr_item, lr_scene_w, lr_scene_h))

        self._im_infos = map(mk_tr_item_image_info, track.tr_track)
        self._im_infos.sort(key=lambda ref: ref.weight, reverse=True)

    def load_measurement(self, def_image_ref):
        """
        :type def_image_ref: models.ImageRef | None
        :rtype: _TrItemImageInfo | None
        """
        self._meas_stats.total += 1
        try:
            im_info = self._vs.export_image_rgba2(self.meas_rgba, self._meas_footer_h, self._im_infos)
            if im_info is not None:
                self._meas_stats.success_arch += 1
                return im_info
        except VsError:
            eventlog.log_w('fail: IVsService::export_image_rgba2')

        if def_image_ref and def_image_ref.url:
            try:
                self._im_loader.load_rgba(self._meas_footer_h, def_image_ref, self.meas_rgba)
                self._meas_stats.success_file += 1
                return None
            except ValueError as x:
                eventlog.log_w('fail: ImLoader::load_rgba (<{}>, {})'.format(def_image_ref, x.message))
            except NotImplementedError:
                eventlog.log_w('fail: ImLoader::load_rgba (<{}>, unsupported format)'.format(def_image_ref))

        self._meas_stats.failure += 1
        raise DetectorError

    def load_overview(self, def_image_ref):
        """
        :type def_image_ref: models.ImageRef | None
        :rtype: _TrItemImageInfo | None
        """
        self._over_stats.total += 1
        if self._over_enable_arch:
            try:
                im_info = self._vs.export_image_jpeg(self.over_jpeg, self._im_infos, '1')
                if im_info is not None:
                    self._over_stats.success_arch += 1
                    return im_info
            except VsError:
                eventlog.log_w('fail: IVsService::export_image_jpeg')

        if def_image_ref and def_image_ref.url:
            try:
                self._im_loader.load_jpeg(def_image_ref, self.over_jpeg)
                self._over_stats.success_file += 1
                return None
            except ValueError as x:
                eventlog.log_w('fail: ImLoader::load_jpeg (<{}>, {})'.format(def_image_ref, x.message))
            except NotImplementedError:
                eventlog.log_w('fail: ImLoader::load_jpeg (<{}>, unsupported format)'.format(def_image_ref))

        self._over_stats.failure += 1
        raise DetectorError


class _ServicePack:
    def __init__(
            self,
            tz_service,
            vs_service,
            raas_service,
            trainer_service,
            about_service,
            lane_detector,
            plate_info_service,
            cam_service,
            clock_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type raas_service: IRaasService
        :type trainer_service: ITrainerService
        :type about_service: IAboutService
        :type lane_detector: LaneDetector
        :type plate_info_service: IPlateInfoService
        :type cam_service: camera.ICameraService | None
        :type clock_service: IClockService
        """
        self.tz = tz_service
        self.vs = vs_service
        self.raas = raas_service
        self.trainer = trainer_service
        self.about = about_service
        self.lane_detector = lane_detector
        self.plate_info = plate_info_service
        self.cam = cam_service
        self.clock = clock_service


class _Detector(IBaseDetector):
    """
    :type _handler: None | (models.Track, BaseEvent) -> None
    """
    def __init__(self, sp, args):
        """
        :type sp: _ServicePack
        :type args: DetectorArgs
        """
        IBaseDetector.__init__(self)

        self._sp = sp
        self._over_im_engine = OverviewImEngine()
        self._meas_im_engine = ImEngine(args.im_engine)
        self._stats = _Stats()

        self._handler = None
        self._def_mapping = args.def_mapping

        self._systest = args.systest
        self._im_provider = _BestImProvider(
            self._sp.vs, args.im_engine.scene.footer.height,
            self._stats.meas_im, self._stats.over_im, args.enable_overview)

    def destroy(self):
        self._im_provider.destroy()

    def set_handler(self, event_handler):
        """
        :type event_handler: (models.Track, BaseEvent) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track):
        """
        :type track: models.Track
        """
        self._im_provider.setup(track)

        if track.wa_track:
            try:
                if track.wa_track is None:
                    wa_best_ref = None
                elif track.wa_track.overview_ref.url is None:
                    wa_best_ref = None
                else:
                    wa_best_ref = track.wa_track.overview_ref

                self._im_provider.load_overview(wa_best_ref)
            except DetectorError:
                eventlog.log_w('fail: _BestImProvider::load_overview')

        try:
            if track.lr_track is None:
                lr_best_ref = None
            elif track.lr_track.best_image.url is None:
                lr_best_ref = None
            else:
                lr_best_ref = track.lr_track.best_image

            best_image_info = self._im_provider.load_measurement(lr_best_ref)
        except DetectorError:
            eventlog.log_e('fail: _BestImProvider::load_measurement')
            raise

        self._analyze(track, best_image_info.tr_item if best_image_info else None)

    def _analyze(self, track, tr_item):
        """
        :type track: models.Track
        :type tr_item: models.TrTrackItem | None
        :rtype: Deferred
        """

        def get_lr_plate(raas_scene):
            """
            :type raas_scene: models.RaasScene
            :rtype: models.Plate
            :raise RaasError:
            """
            raas_plate = raas_scene.try_get_first_scene_item_plate()
            if raas_plate:
                return models.Plate(raas_plate.validity, raas_plate.text, raas_plate.template)

            eventlog.log_d('fail: define raas-result (scene is empty)')
            raise RaasError

        def use_lr_plate(failure):
            """
            :type failure: twisted.python.failure.Failure
            :rtype: models.Plate | None
            """
            # noinspection PyTypeChecker
            failure.trap(RaasError)

            if track.lr_track:
                return track.lr_track.plate

            return None

        def handle_fail(failure):
            """
            :type failure: twisted.python.failure.Failure
            """
            if failure.check([DetectorError]):
                eventlog.log_d('fail: process track')
            else:
                eventlog.log_d('fail: process track (%s)' % str(failure))

        if tr_item:
            rect = tr_item.get_plate_rect(*self._im_provider.meas_rgba.wh)
            quad = tr_item.get_plate_quad(*self._im_provider.meas_rgba.wh)
        else:
            best_img = track.lr_track.best_image
            rect = RectF(*track.lr_track.plate.quad.get_rel_bbox(best_img.width, best_img.height))
            quad = track.lr_track.plate.quad

        # Optimize plate recognition results
        call_chain = self._sp.raas.analyze_scene2(rect, self._im_provider.meas_rgba)
        call_chain.addCallback(get_lr_plate)
        call_chain.addErrback(use_lr_plate)

        # Create an event.
        call_chain.addCallback(self._mk_event_args, track=track, tr_item=tr_item, quad=quad)
        call_chain.addCallback(lambda event_args: self._handler(track, event_args))
        call_chain.addErrback(handle_fail)
        return call_chain

    def _mk_track_item_pos(self, ra_item, lr_item_quad):
        """
        :type ra_item: models.RaTrackItem | None
        :type lr_item_quad: QuadI

        :rtype: Position | None
        """
        if ra_item:
            return Position(ra_item.raw_f, ra_item.raw_l)  # values is cm

        elif self._sp.cam:
            plate_cx, plate_cy = lr_item_quad.center
            world_f, world_l = self._sp.cam.mk_world_point(plate_cx, plate_cy, world_t=0.5)

            # m to cm transform
            world_f = int(world_f * 100)
            world_l = int(world_l * 100)

            # TODO
            world_f = -world_f

            return Position(world_f, world_l)

    def _mk_event_args(self, plate, track, tr_item, quad):
        """
        :type plate: models.Plate | None
        :type track: models.Track
        :type tr_item: models.TrTrackItem | None
        :type quad: QuadI
        :rtype: BaseEvent
        """
        if tr_item:
            wh = self._im_provider.meas_rgba.wh
            wp_direction = track.tr_track.get_item_direction(tr_item)

        else:
            lr_image = track.lr_track.best_image
            wh = lr_image.width, lr_image.height
            wp_direction = Dir(track.lr_track.direction)

        track_direction = track.get_direction()
        wp_dir_alias = self._def_mapping.get(str(wp_direction))

        ts = self._im_provider.meas_rgba.ts

        def mk_vehicle():
            """
            :rtype: Vehicle
            """
            if plate:
                plate_type = LicensePlateType(plate.template, plate.country_a2, self._sp.plate_info)
                plates = [LicensePlate(plate.validity, plate.text, plate_type, quad, track.wa_track)]

            else:
                plates = []

            length = track.ra_track.length if track.ra_track else None
            position = self._mk_track_item_pos(tr_item and tr_item.ra_item, quad)

            ra_speed = track.ra_track.get_median_speed_kmph() if track.ra_track else None
            racam_speed = _mk_racam_speed(track.tr_track)

            if ra_speed and self._systest:
                return Vehicle(
                    plates, length, wp_direction, wp_dir_alias, position,
                    track.lr_track and track.lr_track.type_jnode,
                    ra_speed, racam_speed=None, lpr_speed=None, wa_track=track.wa_track)

            elif ra_speed and racam_speed:
                vehicle_class = 'UNKNOWN'
                if track.lr_track and track.lr_track.type_jnode:
                    vehicle_class = track.lr_track.type_jnode['info'][0]['class']

                rel_error_gt10 = abs(ra_speed - abs(racam_speed)) > 0.1 * ra_speed
                any_car_high_speed = ra_speed >= 90 and rel_error_gt10
                bus_any_speed = vehicle_class in ['UNKNOWN', 'BUS', 'LT', 'HT'] and rel_error_gt10

                if any_car_high_speed or bus_any_speed:
                    if any_car_high_speed:
                        self._stats.ra_cam_speed_invalid_car += 1
                    else:
                        self._stats.ra_cam_speed_invalid_bus += 1

                    return Vehicle(
                        plates, length, wp_direction, wp_dir_alias, position,
                        track.lr_track and track.lr_track.type_jnode,
                        ra_speed=None, racam_speed=None, lpr_speed=None, wa_track=track.wa_track)

                else:
                    return Vehicle(
                        plates, length, wp_direction, wp_dir_alias, position,
                        track.lr_track and track.lr_track.type_jnode,
                        ra_speed, racam_speed, lpr_speed=track.lr_track and track.lr_track.spot_speed,
                        wa_track=track.wa_track)

            else:
                if ra_speed and racam_speed is None:
                    self._stats.ra_cam_speed_absent += 1

                return Vehicle(
                    plates, length, wp_direction, wp_dir_alias, position,
                    track.lr_track and track.lr_track.type_jnode,
                    ra_speed=None, racam_speed=None, lpr_speed=track.lr_track and track.lr_track.spot_speed,
                    wa_track=track.wa_track)

        vehicle = mk_vehicle()
        if track.ra_track is None:
            self._stats.ra_absent += 1

        best_origin = self._sp.about.get_origin(ts)

        def mk_best_location():
            """
            :rtype: Location
            """
            if quad:
                lane_set = self._sp.lane_detector.get_lane3(*quad.center)
                if len(lane_set) == 1:
                    return Location(best_origin.location.place, lane_set.pop())

            return Location(best_origin.location.place)

        self._stats.ra_cam_speed_total += 1

        best_location = mk_best_location()

        scene_dh_data = self._im_provider.meas_rgba
        rect_f = RectF(*quad.get_rel_bbox(*wh))
        plate_bbox = mk_plate_bbox2(rect_f, scene_dh_data.w, scene_dh_data.h)
        control_ts = self._sp.clock.get_control_ts(ts)

        def mk_best_images():
            """
            :rtype: list[MemBlob]
            """
            footer_vars = ImFooterVars(
                self._sp.tz, self._sp.about, ts,
                control_ts=control_ts,
                defs=self._def_mapping,
                location=best_location,
                geolocation=best_origin.geolocation,
                plate_text=vehicle.plate_text,
                spot_speed=vehicle.spot_speed,
                direction=wp_direction,
                radar_speed=vehicle.spot_speed,
                video_speed=vehicle.camera_speed,
                ra_cam_speed=vehicle.ra_cam_speed,
                wim_speed=vehicle.wim_speed,
                main_speed=vehicle.main_speed,
                control_speed=vehicle.control_speed)

            blobs = self._meas_im_engine.mk_blobs(
                scene_dh_data, plate_bbox, footer_vars, track.lr_track, quad, True, True, True, True)

            return blobs

        best_images = mk_best_images()

        if self._im_provider.over_jpeg.ts > 14: # TODO: 14?
            ov_blob = self._over_im_engine.mk_overview(self._im_provider.over_jpeg)
            best_images.append(ov_blob)

        # define #/measurement/0
        measurement = Measurement(track.uuid, direction=track_direction)
        measurement.add_waypoints([
            Waypoint(Waypoint.NAME_CATCH, track.tr_track.min_ts),
            Waypoint(Waypoint.NAME_LOST, track.tr_track.max_ts),
            BestWaypoint(ts, vehicle, control_ts=control_ts, direction=str(vehicle.direction), location=best_location)
                .add_images(best_images)
        ])

        base_event = BaseEvent([measurement], track.lr_track.lr_alarms_jnode if track.lr_track else None)

        def get_blob():
            return self._meas_im_engine.mk_trainer_scene_blob(scene_dh_data)

        self._sp.trainer.collect(base_event, get_blob, rect_f, track)

        return base_event

    def collect_stats(self, json_handler):
        """
        :type json_handler: (dict) -> None
        :rtype: None
        """
        ra_total_fail = self._stats.ra_absent + self._stats.ra_cam_speed_absent +\
            self._stats.ra_cam_speed_invalid_bus + self._stats.ra_cam_speed_invalid_car

        jnode = {
            'type': 'base',
            'name': 'base.0',
            'stat': {
                'load-measurements-image': {
                    'count': [
                        self._stats.meas_im.total,
                        self._stats.meas_im.total - self._stats.meas_im.failure,
                        self._stats.meas_im.failure
                    ],
                    'arch': self._stats.meas_im.success_arch,
                    'file': self._stats.meas_im.success_file
                },
                'load-overview-image': {
                    'count': [
                        self._stats.over_im.total,
                        self._stats.over_im.total - self._stats.over_im.failure,
                        self._stats.over_im.failure
                    ],
                    'arch': self._stats.over_im.success_arch,
                    'file': self._stats.over_im.success_file
                },
                'assert-ra-speed': {
                    'count': [
                        self._stats.ra_cam_speed_total,
                        self._stats.ra_cam_speed_total - ra_total_fail,
                        ra_total_fail
                    ],
                    'ra-speed-absent': self._stats.ra_absent,
                    'racam-speed-absent': self._stats.ra_cam_speed_absent,
                    'racam-speed-invalid-bus': self._stats.ra_cam_speed_invalid_bus,
                    'racam-speed-invalid-car': self._stats.ra_cam_speed_invalid_car
                }
            }
        }

        json_handler(jnode)


def create(
        tz_service,
        vs_service,
        raas_service,
        trainer_service,
        about_service,
        lane_detector,
        plate_info_service,
        cam_service,
        clock_service,
        args):
    """
    :type tz_service: ITzService
    :type vs_service: IVsService
    :type raas_service: IRaasService
    :type trainer_service: ITrainerService
    :type about_service: IAboutService
    :type lane_detector: LaneDetector
    :type plate_info_service: IPlateInfoService
    :type cam_service: camera.ICameraService | None
    :type clock_service: IClockService
    :type args: DetectorArgs

    :rtype: IDetector
    """
    service_pack = _ServicePack(
        tz_service, vs_service, raas_service, trainer_service,
        about_service, lane_detector, plate_info_service, cam_service, clock_service)

    return _Detector(service_pack, args)
