import functools
import math

from edge import eventlog
from edge_bbox.types import mk_plate_bbox2, RectF
from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import FileBlob, ImFooterVars
from edge.service.tz import ITzService
from edge.service.about import IAboutService
from edge.lane_detector import LaneDetector

from .types import DetectorArgs, RedLightEvidence, RuleArgs
from ..types import IDetector
from ..com.types import VideoArgs
from ...model.types import BaseEvent, Event, Waypoint, RedLightViolationWaypoint, BestWaypoint, Vehicle
from ...model.models import Track, TrTrackItem
from ...services.vs import IVsService, VsImageInfo, VsError, VsSessionAbsentError
from ...services.tl import ITlService
from ...services.leddar import ILeddarService


def points_far_enought(rect1, rect2):
    """
    :type rect1: RectF
    :type rect2: RectF
    :rtype: bool
    """
    plate_h = rect1.fb - rect1.ft
    max_distance = plate_h * 0.7  # % of plate height

    rect1_c_x, rect1_c_y = rect1.center
    rect2_c_x, rect2_c_y = rect2.center

    if math.fabs(rect1_c_x - rect2_c_x) < max_distance and math.fabs(rect1_c_y - rect2_c_y) < max_distance:
        # points are in same rect
        return False
    else:
        # points are far enought
        return True


class _SceneImageInfo(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.le_item.ts)
        self.tr_item = tr_item


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, tl_service, leddar_service, lane_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type tl_service: ITlService
        :type leddar_service: ILeddarService
        :type lane_service: LaneDetector
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.tl = tl_service
        self.leddar = leddar_service
        self.lane = lane_service


class Rule:
    def __init__(self, twisted_reactor, services, rule_args, detector_name, postbar_init_im_env,
                 video_conf):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type postbar_init_im_env: ImEnv
        :type video_conf: VideoArgs | None
        """
        self._reactor = twisted_reactor
        self._detector_name = detector_name

        self._services = services
        self._args = rule_args
        self._post_im_env1 = postbar_init_im_env

        self._video_conf = video_conf
        if video_conf:
            self._im_fitter = ImSizeFitter((video_conf.max_w, video_conf.max_h))

    def check(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        :rtype: (Evidence, int) | None
        """
        leddar_result = self._services.leddar.check(track.le_track.min_ts, track.le_track.max_ts)  # TODO
        # print 'le res {}'.format(leddar_result)

        tl_result = self._services.tl.check(track.le_track.min_ts, track.le_track.max_ts)
        # print 'tl- res {}'.format(tl_result)

        # eventlog.log_i("rl2 check {} {}".format(len(tl_result), len(leddar_result)))
        for violation in tl_result:
            for le_violation in leddar_result:
                # prepare im-environments
                post_im_infos = map(_SceneImageInfo, track.tr_track)

                post_im_info1 = self._services.vs.export_image_rgba2(
                    self._post_im_env1.scene_rgba,
                    self._post_im_env1.footer_h,
                    post_im_infos[0:5])  # type: Optional[_SceneImageInfo]
                # prepare evidence
                evidence = self._mk_evidence(
                    base, track,
                    post_im_info1.tr_item)

                if evidence:
                    ref_ts = track.le_track[0].ts / 1000
                    return evidence, ref_ts

    def add_video(self, ref_ts, callback):
        """
        :type ref_ts: float
        :type callback: (FileBlob | None) -> None
        :rtype: None
        """

        if self._video_conf:
            scaled_w, scaled_h = self._im_fitter.scaledown(
                self._post_im_env1.scene_rgba.wh)

            # noinspection PyBroadException
            try:
                return self._services.vs.export_video(
                    ref_ts - self._video_conf.pre_record,
                    ref_ts + self._video_conf.post_record,
                    scaled_w, scaled_h, self._video_conf.quality,
                    callback)

            except VsSessionAbsentError:
                eventlog.log_w('fail: vs (session-absent)')

            except VsError:
                eventlog.log_e('fail: vs')

            except Exception:
                eventlog.log_x()

        callback(None)

    def _mk_evidence(self, base_event, track, post_tr_item1):
        """
        :type base_event: BaseEvent
        :type track: Track
        :type post_tr_item1: TrTrackItem
        :rtype: RedLightEvidence | None
        """
        plate_text = base_event.get_plate_text()

        # get best vehicle
        best_waypoint = base_event.best_wp  # type: BestWaypoint
        best_vehicle = best_waypoint.vehicle  # type: Vehicle

        def mk_waypoint(services, im_env, tr_item, waypoint_name):
            """
            :type services: _RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem
            :type waypoint_name: str
            :rtype: Waypoint
            """
            plate_bbox = post_tr_item1.get_plate_rect()
            origin = services.about.get_origin(im_env.scene_rgba.ts)
            wp_direction = track.tr_track.get_item_direction(tr_item)

            footer_vars = ImFooterVars(services.tz, self._services.about, im_env.scene_rgba.ts,
                                       defs=im_env.footer_def_vars,
                                       location=origin.location, geolocation=origin.geolocation, plate_text=plate_text,
                                       spot_speed=None,
                                       direction=wp_direction,
                                       radar_speed=None, video_speed=None,
                                       ra_cam_speed=None)

            im_blobs = im_env.engine.mk_blobs(
                im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.get_plate_quad(*im_env.scene_rgba.wh),
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_d('fail: mk image-blobs')
                raise RuntimeError

            waypoint = RedLightViolationWaypoint(
                waypoint_name, im_env.scene_rgba.ts, origin.location, [tr_item.get_plate_quad(*im_env.scene_rgba.wh)],
                *services.tl.get_times(im_env.scene_rgba.ts))

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            post_waypoint1 = mk_waypoint(
                self._services, self._post_im_env1, post_tr_item1,
                Waypoint.NAME_POST_STOPBAR)
        except RuntimeError as x:
            eventlog.log_x()
            return None

        event_ts = self._post_im_env1.scene_rgba.ts

        evidence = RedLightEvidence(
            self._services.about.get_origin(event_ts),
            event_ts, self._args, self._detector_name)
        evidence.add_waypoints([post_waypoint1])
        return evidence


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self, twisted_reactor, tz_service, vs_service, about_service, tl_service, leddar_service, lane_service, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type vs_service: IVsService
        :type tz_service: ITzService
        :type tl_service: ITlService
        :type leddar_service: ILeddarService
        :type lane_service: LaneDetector
        :type args: DetectorArgs
        """
        IDetector.__init__(self)

        im_engine = ImEngine(args.im_engine)
        postbar_init_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        self._rules = []

        self._le = leddar_service
        rule_services = _RuleServices(tz_service, vs_service, about_service, tl_service, leddar_service, lane_service)
        for rule in args.rules:
            rule = Rule(
                twisted_reactor,
                rule_services, rule, args.name, postbar_init_im_env,
                args.video)
            self._rules.append(rule)

        self._identity = args.identity
        self._handler = None

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def handle_evidence(self, base, evidence, video_blob):
        """
        :type base: BaseEvent
        :type evidence: RedLightEvidence
        :type video_blob: FileBlob | None
        """
        if video_blob:
            evidence.add_videos([video_blob])

        self._handler(Event(self._identity, base, evidence))

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        # print 'rl2-analyze {} {} {}'.format(track.le_track.ref, track.le_track.min_ts, track.le_track.max_ts)
        if track.le_track and track.le_track.ref != self._le.ref():
            return

        for rule in self._rules:
            reply = rule.check(track, base)
            if reply is None:
                continue

            evidence, ref_ts = reply
            if evidence:
                handler = functools.partial(self.handle_evidence, base, evidence)
                rule.add_video(ref_ts, handler)
                break
