# coding=utf-8
import base64
import ujson
import urllib2

from twisted.internet import reactor
from twisted.internet.defer import DeferredList, inlineCallbacks, returnValue
from twisted.web.client import Agent, readBody
from twisted.web.http_headers import Headers

from edge import eventlog
from edge.image.image import ImEnv, ImEngine
from edge.image.types import ImFooterVars, MemBlob, JsonBlobMeta
from edge.service.tz import isostr_to_timestamp_usec
from edge_bbox.types import mk_plate_bbox2, RectF, QuadI

from .types import SectionSpeedEvidence, SectionArgs, DetectorArgs, RuleArgs
from ..com.schedule import Schedule
from ..types import IDetector
from ...model.models import Track, TrTrackItem
from ...model.types import Event, Waypoint, BaseEvent, Position
from ...services.vs import VsImageInfo
from ...services.clock import IClockService


class BaseEvent2(BaseEvent):
    def __init__(self, base_event, remote_m0):
        """
        :type base_event: BaseEvent

        :param remote_m0: remote-args.json#/measurements/0
        :type remote_m0: dict
        """
        BaseEvent.__init__(self, base_event.measurements, base_event.alarms_jnode)
        self._measurement_jnode = remote_m0

    def serialize(self, tz_service):
        base_jnode = BaseEvent.serialize(self, tz_service)
        base_jnode['measurements'].append(self._measurement_jnode)
        return base_jnode


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, clock_service, cam_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type clock_service: IClockService
        :type cam_service: camera.ICameraService | None
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.clock = clock_service
        self.cam_service = cam_service


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.ts)
        self.tr_item = tr_item


class Rule:
    def __init__(self, services, rule_args, detector_name, section_args, ignore_ra,
                 fini_im_env):
        """
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type section_args: SectionArgs
        :type ignore_ra: bool
        :type fini_im_env: ImEnv
        """
        self._detector_name = detector_name

        self._services = services
        self._scheduler = Schedule(rule_args.schedule, services.tz)
        self._rule_conf = rule_args
        self._det_conf = section_args
        self._fini_im_env = fini_im_env
        self._ignore_ra = ignore_ra

        self._agent = Agent(reactor, connectTimeout=60)

    @inlineCallbacks
    def _get_remote_event_file(self, event_id, img_name):
        """
        :type event_id: str
        :type img_name: str
        :rtype: bytes
        """
        det_conf = self._det_conf
        user, password = det_conf.init_area.auth
        base_url = det_conf.init_area.controller

        url = '{url}/api/v2/origins/event/dist/{event_id}/by-ref/{img_name}'.format(
            url=base_url, event_id=event_id, img_name=img_name)
        b64 = base64.b64encode('{}:{}'.format(user, password))
        resp = yield self._agent.request(b'GET', url, Headers({
            'Authorization': ['Basic {}'.format(b64)]
        }), None)

        if resp.code != 200:
            eventlog.log_e('fail: get file: <{0}>, <{1}>'.format(url, resp.code))
            returnValue(None)

        file_contents = yield readBody(resp)
        returnValue(file_contents)

    @inlineCallbacks
    def _get_remote_feed_record(self, plate_text):
        """
        :type plate_text: str
        :rtype: dict | None
        """
        det_conf = self._det_conf
        user, password = det_conf.init_area.auth
        base_url = det_conf.init_area.controller
        origin = det_conf.init_area.origin
        direction = det_conf.direction

        url = (u'{url}/api/v2/origins/event/feed?origin={origin}&type=transit&'
               u'dir={dir}&position=required&plate-method=exact&plate-pattern={plate}&limit=1'
               ).format(url=base_url, origin=origin, dir=direction, plate=plate_text)

        url = urllib2.quote(url.encode('utf8'), ':/&?=')
        b64 = base64.b64encode('{}:{}'.format(user, password))
        resp = yield self._agent.request(b'GET', url, Headers({
            "Authorization": ["Basic {}".format(b64)]
        }), None)

        if resp.code != 200:
            eventlog.log_e('fail: get feed record: <{0}>, <{1}>'.format(url, resp.code))
            returnValue(None)

        feed_str = yield readBody(resp)
        feed_records = ujson.loads(feed_str)  # list of feed records
        if feed_records:
            returnValue(feed_records[0])

        else:
            returnValue(None)

    def _estimate_speed_dist(self, remote_ts, remote_dist, local_ts, local_dist):
        """
        :param remote_ts: time in usec
        :type remote_ts: int

        :param remote_dist: init distance in cm
        :type remote_dist: int

        :param local_ts: time in usec
        :type local_ts: int

        :param local_dist: fini distance in cm
        :type local_dist: int

        :rtype: (float, float)
        """
        conf_dir, conf_dist = self._det_conf.direction, self._det_conf.dist
        if conf_dir.dir == conf_dir.RECEDING:
            dist_m = (conf_dist - remote_dist / 100.0 + local_dist / 100.0)
        else:
            dist_m = (conf_dist + remote_dist / 100.0 - local_dist / 100.0)

        time_hours = (local_ts - remote_ts) / 1e6 / 3600
        dist_km = dist_m / 1000.0

        section_speed = dist_km / time_hours
        return section_speed, dist_m

    def _estimate_speed(self, remote_record, local_track, base):
        """
        :type remote_record: dict
        :type local_track: Track
        :type base: BaseEvent
        :rtype: float
        """
        remote_e = remote_record['event']
        remote_v = remote_e['vehicle']
        remote_p = remote_v.get('position') or dict()

        remote_ts = isostr_to_timestamp_usec(remote_e['time'])  # usec
        remote_dist = remote_p.get('forward', 0)  # cm

        conf_dir, conf_dist = self._det_conf.direction, self._det_conf.dist

        local_best_wp = base.measurements[0].best_waypoint
        local_best_position = local_best_wp.vehicle.position

        if conf_dist < self._det_conf.asssert_dist and not self._ignore_ra:
            assert local_track.ra_track
            local_ts = local_track.ra_track[0].ts
            local_dist = local_track.ra_track[0].raw_f

        elif local_track.ra_track and not self._ignore_ra:
            # TODO: почему-бы не брать best ?
            local_ts = local_track.ra_track[0].ts
            local_dist = local_track.ra_track[0].raw_f

        elif local_best_position is not None:
            local_ts = local_best_wp.ts
            local_dist = local_best_position.forward

        elif conf_dir.dir == conf_dir.APPROACHING:
            local_ts = local_track.lr_track.max_ts
            local_dist = 0

        elif conf_dir.dir == conf_dir.RECEDING:
            local_ts = local_track.lr_track.min_ts
            local_dist = 0

        else:
            local_ts = local_best_wp.ts
            local_dist = 0

        speed, _ = self._estimate_speed_dist(remote_ts, remote_dist, local_ts, local_dist)
        return speed

    @inlineCallbacks
    def check(self, local_track, base):
        """
        :type local_track: Track
        :type base: BaseEvent
        :rtype: SectionSpeedEvidence | None
        """
        if self._det_conf.dist < self._det_conf.asssert_dist and local_track.ra_track is None:
            eventlog.log_w('fail: radar measurements are required')
            returnValue(None)
            return

        min_ts = local_track.lr_track.min_ts
        max_ts = local_track.lr_track.max_ts
        if not(self._scheduler.check2(max_ts) or self._scheduler.check2(min_ts)):
            returnValue(None)

        plate_text = base.get_plate_text()
        try:
            feed_record = yield self._get_remote_feed_record(plate_text)
        except Exception:
            eventlog.log_e('fail: get remote feed record')
            returnValue(None)
            return

        if feed_record is None:  # !200 or empty list
            returnValue(None)
            return

        try:
            est_section_speed = self._estimate_speed(feed_record, local_track, base)
        except KeyError:
            returnValue(None)
            return

        conf_speed_limit = self._rule_conf.limits.speed_limit
        conf_max_overspeed = self._rule_conf.limits.over_speed
        conf_max_section_speed = conf_speed_limit + conf_max_overspeed
        eventlog.log_d('plate:{}, speed:{}, max-speed:{}'.format(plate_text.encode('utf-8'), est_section_speed, conf_max_section_speed))
        if est_section_speed <= conf_max_section_speed:
            returnValue(None)
            return

        # download args.json and images
        remote_e = feed_record['event']
        try:
            remote_uuid = remote_e['uuid']
            remote_args_name = feed_record['dist']['args']
            remote_args_bytes = yield self._get_remote_event_file(remote_uuid, remote_args_name)
            remote_args = ujson.loads(remote_args_bytes)

            remote_m0 = remote_args['measurements'][0]
            remote_best = remote_m0['waypoints']['best']
            remote_v = remote_best['vehicle']
            remote_plate_quad = QuadI.unpack(remote_v['license-plates'][0]['bbox-4'])

            remote_i = remote_best['images']
            remote_scene_imname = remote_i['scene']['name']
            remote_plate_imname = remote_i['license-plates'][0]['name']

            remote_scene_iminfo = remote_i['scene']['info']

            remote_scene_imbytes = yield self._get_remote_event_file(remote_uuid, remote_scene_imname)
            remote_plate_imbytes = yield self._get_remote_event_file(remote_uuid, remote_plate_imname)

        except Exception:
            returnValue(None)
            return

        if local_track.ra_track is not None:
            candidate_items = filter(lambda x: x.lr_item and x.ra_item, local_track.tr_track)  # type: list[TrTrackItem]
        else:
            candidate_items = local_track.tr_track

        scene_ms_refs = map(_SceneMsRef, candidate_items)

        local_scene_iminfo = self._services.vs.export_image_rgba2(
            self._fini_im_env.scene_rgba,
            self._fini_im_env.footer_h,
            scene_ms_refs[:-6:-1])  # type: Optional[_SceneMsRef]
        if not local_scene_iminfo:
            eventlog.log_w('fail: get local_scene_im from video storage')
            returnValue(None)
            return

        local_ra_item = local_scene_iminfo.tr_item.ra_item
        local_cam_f = 0
        local_cam_l = 0
        # если нет радара и камера MV, пересчитать forward через image-mapper
        if self._services.cam_service:
            plate_cx, plate_cy = local_scene_iminfo.tr_item.lr_item.plate_quad.center
            world_f, world_l = self._services.cam_service.mk_world_point(plate_cx, plate_cy, world_t=0.5)
            # m to cm transform

            local_cam_l = int(world_l * 100)
            local_cam_f = - int(world_f * 100)

        if local_ra_item and not self._ignore_ra:
            local_f = local_ra_item.raw_f
            local_l = local_ra_item.raw_l
        else:
            local_l = local_cam_l
            local_f = local_cam_f

        local_pos = None if local_f == local_l == 0 else Position(local_f, local_l)
        local_ts = local_scene_iminfo.tr_item.ts

        remote_p = remote_e['vehicle'].get('position') or dict()
        remote_f = remote_p.get('forward', 0)  # cm
        remote_l = remote_p.get('left', 0)  # cm
        remote_pos = Position(remote_f, remote_l) if remote_p else None
        remote_ts = isostr_to_timestamp_usec(remote_e['time'])  # usec

        section_speed, section_dist = self._estimate_speed_dist(remote_ts, remote_f, local_ts, local_f)
        section_speed = round(section_speed, 1)

        if local_ra_item:
            eventlog.log_d('plate:{}, remote_f:{:6}, remote_ts: {}, local_f:{:6}, local_ra_f:{:6}, local_ts: {}, speed:{:6}, max-speed:{:6}'.format(
                plate_text,
                remote_f,
                remote_ts,
                local_f,
                local_ra_item.raw_f,
                local_ts,
                section_speed,
                conf_max_section_speed))
        else:
            eventlog.log_d('plate:{}, remote_f:{:6}, remote_ts: {}, local_f:{:6}, local_ts: {}, speed:{:6}, max-speed:{:6}'.format(
                plate_text,
                remote_f,
                remote_ts,
                local_f,
                local_ts,
                section_speed,
                conf_max_section_speed))

        if section_speed <= conf_max_section_speed:
            returnValue(None)
            return

        evidence = self._mk_evidence(plate_text, section_dist, section_speed, remote_best, remote_scene_imbytes,
                                     remote_scene_iminfo, remote_plate_quad, remote_plate_imbytes, remote_pos,
                                     local_track, local_scene_iminfo.tr_item, local_pos)

        del remote_best['images']
        remote_m0['origin'] = remote_args['origin']
        remote_args_crt = remote_args.get('certificate')
        if remote_args_crt:
            remote_m0['certificate'] = remote_args_crt

        # remote_m0['transit-uuid'] = args_json['uuid']
        returnValue((evidence, remote_m0))

    def _mk_evidence(self, plate_text, section_dist, section_speed, remote_best, remote_scene_imbytes,
                     remote_scene_iminfo, remote_plate_quad, remote_plate_imbytes, remote_pos, local_track,
                     local_tr_item, local_pos):
        """
        :type plate_text: str
        :type section_dist: float
        :type section_speed: float

        :type remote_best: dict
        :type remote_scene_imbytes: bytes
        :type remote_scene_iminfo: dict
        :type remote_plate_quad : QuadI
        :type remote_plate_imbytes: bytes
        :type remote_pos: Position | None

        :type local_track: Track
        :type local_tr_item: TrTrackItem
        :type local_pos: Position | None

        :rtype: SectionSpeedEvidence | None
        """

        def mk_vehicle_jnode(quad, position=None):
            """
            :type quad: QuadI
            :type position: Position | None
            :rtype: dict
            """
            jnode = {
                'vehicle': {
                    'license-plates': [{
                        'bbox-4': quad.serialize(),
                        'bbox-2': quad.get_abs_bbox()
                    }]
                }
            }
            if position:
                jnode['vehicle']['position'] = position.serialize()

            return jnode

        def create_remote_wp():
            """
            :rtype: Waypoint
            """
            remote_ts = isostr_to_timestamp_usec(remote_best['time'])
            remote_control_ts_str = remote_best.get('control-time')
            remote_control_ts = isostr_to_timestamp_usec(remote_control_ts_str) if remote_control_ts_str else None
            answer = Waypoint('init', remote_ts, control_ts=remote_control_ts,
                              draft_jnode=mk_vehicle_jnode(remote_plate_quad, remote_pos))

            scene = MemBlob('scene', remote_ts, bytes(remote_scene_imbytes), meta=JsonBlobMeta(remote_scene_iminfo))
            plate = MemBlob('plate', remote_ts, bytes(remote_plate_imbytes))
            answer.add_images([scene, plate])
            return answer

        remote_wp = create_remote_wp()

        local_plate_quad = local_tr_item.lr_item.plate_quad
        local_plate_rect = RectF(*local_plate_quad.get_rel_bbox(
            local_track.lr_track.best_image.width,
            local_track.lr_track.best_image.height))
        plate_bbox = mk_plate_bbox2(local_plate_rect, *self._fini_im_env.scene_rgba.wh)

        waypoint_ts = self._fini_im_env.scene_rgba.ts
        control_ts = self._services.clock.get_control_ts(waypoint_ts)

        waypoint_origin = self._services.about.get_origin(waypoint_ts)
        footer_vars = ImFooterVars(self._services.tz, self._services.about,
                                   waypoint_ts,
                                   control_ts=control_ts,
                                   defs=self._fini_im_env.footer_def_vars,
                                   location=waypoint_origin.location,
                                   geolocation=waypoint_origin.geolocation,
                                   plate_text=plate_text,
                                   section_speed=section_speed,
                                   section_overspeed=section_speed - self._rule_conf.limits.speed_limit,
                                   direction=self._det_conf.direction,
                                   max_speed=self._rule_conf.limits.speed_limit)

        im_blobs = self._fini_im_env.engine.mk_blobs(
            self._fini_im_env.scene_rgba, plate_bbox, footer_vars, local_track.lr_track,
            local_plate_quad, scene=True, plate=True)

        if len(im_blobs) != 2:
            eventlog.log_d('fail: get all blobs for evidence')
            raise RuntimeError

        local_wp = Waypoint('fini', waypoint_ts, control_ts=control_ts,
                            draft_jnode=mk_vehicle_jnode(local_plate_quad, local_pos))
        local_wp.add_images(im_blobs)

        evidence = SectionSpeedEvidence(
            self._services.about.get_origin(waypoint_ts), waypoint_ts,
            section_speed, self._rule_conf, self._detector_name, self._det_conf.direction,
            self._det_conf.dist, section_dist, control_ts=control_ts)
        evidence.add_waypoints([remote_wp, local_wp])
        return evidence


class _Detector(IDetector):
    def __init__(self, tz_service, vs_service, about_service, clock_service, cam_service, args):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type clock_service: IClockService
        :type cam_service: camera.ICameraService | None
        :type args: DetectorArgs
        """
        self._identity = args.identity

        self._handler = None

        self._fini_im_env = ImEnv(
            ImEngine(args.im_engine),
            args.im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._rules = []
        rule_services = _RuleServices(tz_service, vs_service, about_service, clock_service, cam_service)
        for rule in args.rules:
            if rule.enabled:
                self._rules.append(Rule(rule_services, rule,
                                        args.name, args.section, args.ignore_ra,
                                        self._fini_im_env))
        self._section = args.section

        IDetector.__init__(self)

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        def log_exc(e):
            eventlog.log_w(str(e))

        def handle_results(d_results):
            """
            :param d_results: list of pairs (deferred success, deferred result)
            :type d_results: list (bool, (SectionSpeedEvidence, dict) | None)
            """
            for success, reply in d_results:
                if success and reply:
                    (evidence, jnode) = reply
                    self._handler(Event(self._identity, BaseEvent2(base, jnode), evidence))

        if track.lr_track.direction != self._section.direction.dir:
            return

        if '*' in base.get_plate_text():
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        deferreds = []
        for rule in self._rules:
            deferreds.append(rule.check(track, base))

        d_list = DeferredList(deferreds)
        d_list.addCallbacks(handle_results, log_exc)

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler
