import functools

from edge import eventlog
from edge_bbox.types import mk_plate_bbox2, RectF
from edge.lane_detector import LaneDetector
from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import FileBlob, ImFooterVars
from edge.service.tz import ITzService
from edge.model.types import Dir
from edge.service.about import IAboutService
from edge.landmark import camera

from .types import DetectorArgs, RuleArgs, SpotSpeedWagaEvidence
from ..base.best import check_quad_dist_from_img_borders
from ..types import IDetector
from ..com.schedule import Schedule
from ..com.types import VideoArgs
from ...model.models import Track, TrTrackItem, LrTrackItem
from ...model.types import Event, Position, ViolationWaypoint, BaseEvent, Vehicle, Waypoint, BestViolationWaypoint
from ...services.vs import IVsService, VsImageInfo, VsError, VsSessionAbsentError


def _mk_cam_speed(cam_service, min_lr_item, max_lr_item, plate_up=0.5):
    """
    :type cam_service: camera.ICameraService
    :type min_lr_item: LrTrackItem
    :type max_lr_item: LrTrackItem
    :type plate_up: float

    :return: Speed measured by camera. Return 0 on error.
    :rtype: float
    """
    if min_lr_item.ts == max_lr_item.ts:
        return 0

    success, cam_speed = cam_service.get_speed(
        plate_up,
        min_lr_item.plate_quad,
        max_lr_item.plate_quad,
        min_lr_item.ts,
        max_lr_item.ts)

    return cam_speed if success else 0


class _TrItemImageInfo(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.ts)
        self.tr_item = tr_item


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, lane_detector, cam_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type lane_detector: LaneDetector
        :type cam_service: camera.ICameraService | None
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.lane = lane_detector
        self.cam_service = cam_service


class _RuleStats:
    def __init__(self):
        self.total = 0
        self.schedule = 0
        self.lane = 0
        self.vehicle_type = 0
        self.valid_speed = 0
        self.cam_speed = 0

        self.make_evidence_total = 0
        self.make_evidence_load_images = 0
        self.make_evidence_make_images = 0


class _Rule:
    def __init__(self, twisted_reactor, services, rule_args, detector_name, stats_name,
                 init_im_env, fini_im_env, best_im_env, video_conf, h_over_ground=None, log=False):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type stats_name: str
        :type init_im_env: ImEnv
        :type fini_im_env: ImEnv
        :type best_im_env: ImEnv
        :type video_conf: VideoArgs | None
        :type h_over_ground: float | None
        :type log: bool
        """
        self._stats_name = stats_name
        self._h_over_ground = h_over_ground
        self._log = log

        self._reactor = twisted_reactor
        self._detector_name = detector_name

        self._init_im_env = init_im_env
        self._fini_im_env = fini_im_env
        self._best_im_env = best_im_env

        self._services = services
        self._scheduler = Schedule(rule_args.schedule, services.tz)
        self._args = rule_args

        self._stats = _RuleStats()

        self._video_conf = video_conf
        if video_conf:
            self._im_fitter = ImSizeFitter((video_conf.max_w, video_conf.max_h))

    def check(self, track, base, lane_to_tr_item_list):
        """
        :type track: Track
        :type base: BaseEvent
        :type lane_to_tr_item_list: list of (str, TrTrackItem)
        :rtype: SpotSpeedWagaEvidence | None
        """
        self._stats.total += 1
        plate_text_utf8 = base.get_plate_text().encode('utf-8')

        # workaround
        best_vehicle = base.best_wp.vehicle  # type: Vehicle

        if best_vehicle.vehicle_class not in self._args.limits.vehicle_types:
            self._stats.vehicle_type += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'type-unmatch'))
            return None

        if not (self._scheduler.check2(track.lr_track.max_ts) or self._scheduler.check2(track.lr_track.min_ts)):
            self._stats.schedule += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'schedule-unmatch'))
            return None

        tr_items = []
        for lane, tr_item in lane_to_tr_item_list:
            if lane in self._args.limits.lanes or len(self._args.limits.lanes) == 0:
                if tr_item.lr_item:
                    lr_scene_img = track.lr_track.best_image
                    # filter lr track near to img edge
                    if check_quad_dist_from_img_borders(
                            tr_item.lr_item.plate_quad, lr_scene_img.width, lr_scene_img.height):
                        tr_items.append(tr_item)

        if len(tr_items) < 2:
            self._stats.lane += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'lane-unmatch'))
            return None

        tr_items.sort(key=lambda ref: ref.ts)

        wa_speed = track.wa_track.speed
        max_speed = self._args.limits.speed_limit + self._args.limits.over_speed

        # get videospeed
        if self._h_over_ground is not None:
            cam_speed = _mk_cam_speed(
                self._services.cam_service, tr_items[0].lr_item, tr_items[-1].lr_item, self._h_over_ground)

            if self._log:
                eventlog.log_i(
                    'camspeed {%6.3f} waspeed{ %3f} epsilon {%7.3f} text {%s}' %
                    (cam_speed, wa_speed, wa_speed - cam_speed, plate_text_utf8))

            rel_error_gt10 = abs(cam_speed - wa_speed) > 0.1 * wa_speed
            any_car_high_speed = wa_speed >= 90 and rel_error_gt10
            bus_any_speed = best_vehicle.vehicle_class in ['UNKNOWN', 'BUS', 'LT', 'HT'] and rel_error_gt10

            if cam_speed == 0 or any_car_high_speed or bus_any_speed:
                eventlog.log_i('fail: reject track (plate:{}, reason:{}, camspeed {%6.3f} wa-speed{ %3f})'.format(
                    plate_text_utf8, 'cam-speed-invalid', cam_speed, wa_speed))

                self._stats.cam_speed += 1
                return None

        max_error = 2 if wa_speed < 100 else (wa_speed * 0.02)
        if wa_speed <= max_speed + max_error:
            self._stats.valid_speed += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'wa-speed-valid'))
            return None

        self._stats.make_evidence_total += 1

        # overspeed
        image_infos = map(_TrItemImageInfo, tr_items)
        init_image_infos = image_infos[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)

        fini_image_infos = image_infos[:-6:-1]
        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        best_image_infos = image_infos #TODO

        init_image_info = self._services.vs.export_image_rgba2(
            self._init_im_env.scene_rgba, self._init_im_env.footer_h, init_image_infos)

        best_image_info = self._services.vs.export_image_rgba2(
            self._best_im_env.scene_rgba, self._best_im_env.footer_h, best_image_infos)

        fini_image_info = self._services.vs.export_image_rgba2(
            self._fini_im_env.scene_rgba, self._fini_im_env.footer_h, fini_image_infos)

        if None in [init_image_info, fini_image_info, best_image_info]:
            self._stats.make_evidence_load_images += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'load-images'))
            return None

        return self._mk_evidence(
            base, track, wa_speed, best_vehicle.camera_speed, self._args.limits.speed_limit,
            init_image_info.tr_item,
            fini_image_info.tr_item,
            best_image_info.tr_item)

    def _mk_evidence(self, base_event, track, wim_speed, cam_speed, speed_limit,
                     init_tr_item, fini_tr_item, best_tr_item):
        """
        :type base_event: BaseEvent
        :type track: Track
        :type wim_speed: int
        :type cam_speed: int
        :type speed_limit: int
        :type init_tr_item: TrTrackItem
        :type fini_tr_item: TrTrackItem
        :type best_tr_item: TrTrackItem
        :rtype: SpotSpeedWagaEvidence | None
        """
        plate_text = base_event.get_plate_text()
        plate_text_utf8 = plate_text.encode('utf-8')

        # get best vehicle
        best_vehicle = base_event.best_wp.vehicle  # type: Vehicle

        def mk_waypoint(services, im_env, tr_item, w_type=Waypoint.NAME_VIOLATION):
            """
            :type services: _RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem
            :type w_type: str
            :rtype: ViolationWaypoint | BestViolationWaypoint
            """
            plate_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(
                track.lr_track.best_image.width,
                track.lr_track.best_image.height))
            plate_bbox = mk_plate_bbox2(plate_rect, *im_env.scene_rgba.wh)

            waypoint_ts = im_env.scene_rgba.ts
            # waypoint_pos = Position(tr_item.ra_item.raw_f, tr_item.ra_item.raw_l)
            waypoint_pos = Position(14, 14) # TODO
            wp_dir = track.tr_track.get_item_direction(tr_item)

            wp_dir_alias = im_env.footer_def_vars.get(str(wp_dir))

            waypoint_origin = services.about.get_origin(waypoint_ts)
            lane_set = self._services.lane.get_lane3(*tr_item.get_plate_quad().center)
            if len(lane_set) == 1:
                waypoint_origin.location.lane = lane_set.pop()

            footer_vars = ImFooterVars(
                services.tz, self._services.about, waypoint_ts, defs=im_env.footer_def_vars,
                location=waypoint_origin.location,
                geolocation=waypoint_origin.geolocation,
                plate_text=plate_text,
                spot_speed=wim_speed,
                direction=wp_dir,
                video_speed=cam_speed,
                radar_speed=wim_speed,
                max_speed=speed_limit,
                wim_overspeed=wim_speed - speed_limit,
                main_overspeed=wim_speed - speed_limit,
                ra_cam_speed=best_vehicle.ra_cam_speed,
                wim_speed=best_vehicle.wim_speed,
                main_speed=best_vehicle.main_speed,
                control_speed=best_vehicle.control_speed)

            waypoint = None
            if w_type == Waypoint.NAME_VIOLATION:
                im_blobs = im_env.engine.mk_blobs(
                    im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True)

                if len(im_blobs) != 3:
                    eventlog.log_w('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint = ViolationWaypoint(
                    waypoint_ts, waypoint_origin.location,
                    direction=wp_dir,
                    direction_alias=wp_dir_alias,
                    position=waypoint_pos,
                    plate_quads=[tr_item.lr_item.plate_quad])

            elif w_type == Waypoint.NAME_VIOLATION_BEST:
                im_blobs = im_env.engine.mk_blobs(
                    im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True, thumb=True)

                if len(im_blobs) != 4:
                    eventlog.log_w('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint = BestViolationWaypoint(
                    waypoint_ts, waypoint_origin.location,
                    direction=wp_dir,
                    direction_alias=wp_dir_alias,
                    position=waypoint_pos,
                    plate_quads=[tr_item.lr_item.plate_quad])
            else:
                eventlog.log_e('fail: unsupported')
                raise RuntimeError

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            init_wp = mk_waypoint(self._services, self._init_im_env, init_tr_item)
            fini_wp = mk_waypoint(self._services, self._fini_im_env, fini_tr_item)
            best_wp = mk_waypoint(self._services, self._best_im_env, best_tr_item, Waypoint.NAME_VIOLATION_BEST)
        except RuntimeError:
            self._stats.make_evidence_make_images += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'make-images'))
            return None

        event_ts = self._fini_im_env.scene_rgba.ts
        direction = Dir(Dir.UNKNOWN)
        # TODO
        # init_position = Position(init_tr_item.ra_item.raw_f, init_tr_item.ra_item.raw_l)
        # fini_position = Position(fini_tr_item.ra_item.raw_f, fini_tr_item.ra_item.raw_l)
        # if init_position.forward > fini_position.forward:
        #     direction = Dir(Dir.APPROACHING)
        #
        # elif init_position.forward < fini_position.forward:
        #     direction = Dir(Dir.RECEDING)

        evidence = SpotSpeedWagaEvidence(
            self._services.about.get_origin(event_ts), event_ts, wim_speed,
            cam_speed, self._args, self._detector_name, direction)

        evidence.add_waypoints([init_wp, best_wp, fini_wp])
        return evidence

    def add_video(self, handler):
        """
        :type handler: (FileBlob | None) -> None
        :rtype: None
        """
        if self._video_conf:
            scaled_w, scaled_h = self._im_fitter.scaledown(self._init_im_env.scene_rgba.wh)

            # noinspection PyBroadException
            try:
                return self._services.vs.export_video(
                    self._init_im_env.scene_rgba.ts / 1000 - self._video_conf.pre_record,
                    self._fini_im_env.scene_rgba.ts / 1000 + self._video_conf.post_record,
                    scaled_w, scaled_h, self._video_conf.quality,
                    handler)

            except VsSessionAbsentError:
                eventlog.log_w('fail: VsSessionAbsentError')

            except VsError:
                eventlog.log_w('fail: VsError')

            except Exception:
                eventlog.log_x()

        handler(None)

    def collect_stats(self, json_handler):
        """
        :type json_handler: (dict) -> None
        :rtype: None
        """
        make_evidence_total_fail = self._stats.make_evidence_make_images + self._stats.make_evidence_load_images
        total_fail = self._stats.schedule + self._stats.lane + self._stats.valid_speed + \
                     self._stats.cam_speed + self._stats.vehicle_type
        jnode = {
            'type': 'spot-speed-waga~rule',
            'name': self._stats_name,
            'stat': {
                'assert-track': {
                    'count': [
                        self._stats.total,
                        self._stats.total - total_fail,
                        total_fail
                    ],
                    'type-unmatch': self._stats.vehicle_type,
                    'schedule-unmatch': self._stats.schedule,
                    'lane-unmatch': self._stats.lane,
                    'ra-speed-valid': self._stats.valid_speed,
                    'cam-speed-invalid': self._stats.cam_speed
                },
                'make-evidence': {
                    'count': [
                        self._stats.make_evidence_total,
                        self._stats.make_evidence_total - make_evidence_total_fail,
                        make_evidence_total_fail
                    ],
                    'load-images': self._stats.make_evidence_load_images,
                    'make-images': self._stats.make_evidence_make_images
                }
            }
        }

        json_handler(jnode)


class _DetectorStats:
    def __init__(self):
        self.total = 0
        self.wa_absent = 0
        self.lr_absent = 0
        self.lr_invalid = 0

        self.emit_total = 0
        self.emit_fail = 0
        self.ra_cam_speed = 0


class _Detector(IDetector):
    def __init__(self, twisted_reactor, tz_service, vs_service, about_service, lane_detector, cam_service, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type lane_detector: LaneDetector
        :type cam_service: camera.ICameraService | None
        :type args: DetectorArgs
        """
        self._stats = _DetectorStats()
        self._identity = args.identity
        self._lane_detector = lane_detector

        self._handler = None

        self._init_im_env = ImEnv(
            ImEngine(args.im_engine),
            args.im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._fini_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._best_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._rules = []  # type: list[_Rule]
        rule_services = _RuleServices(tz_service, vs_service, about_service, lane_detector, cam_service)
        rule_index = 0
        for rule_conf in args.rules:
            if rule_conf.enabled:  # TODO: strip disabled rules when application starts
                rule_name = 'spot-speed-waga.{}~rule.{}'.format(self._identity.id, rule_index)
                rule_index += 1
                rule = _Rule(
                    twisted_reactor, rule_services, rule_conf, args.name, rule_name,
                    self._init_im_env, self._fini_im_env, self._best_im_env,
                    args.video, args.h_over_ground, args.log)

                self._rules.append(rule)

        IDetector.__init__(self)

    def handle_evidence(self, base, evidence, video_blob):
        """
        :type base: BaseEvent
        :type evidence: SpotSpeedWagaEvidence
        :type video_blob: FileBlob | None
        """
        if video_blob:
            evidence.add_videos([video_blob])

        self._stats.emit_total += 1

        if not self._handler(Event(self._identity, base, evidence)):
            self._stats.emit_fail += 1

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        self._stats.total += 1
        if track.lr_track is None:
            self._stats.lr_absent += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                '-', 'lr-absent'))
            return

        plate_text_utf8 = base.get_plate_text().encode('utf-8')

        if track.wa_track is None:
            self._stats.wa_absent += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'wa-absent'))
            return

        # best_vehicle = base.best_wp.vehicle  # type: # Vehicle
        # base_racam_speed = best_vehicle.ra_cam_speed
        #
        # if base_racam_speed is None:
        #     self._stats.ra_cam_speed += 1
        #     eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
        #         plate_text_utf8, 'racam-speed-invalid'))
        #     return

        if '*' in base.get_plate_text():
            self._stats.lr_invalid += 1
            eventlog.log_i('fail: reject track (plate:{}, reason:{})'.format(
                plate_text_utf8, 'lr-invalid'))
            return

        def get_parts():
            """
            :rtype: list of (str, TrTrackItem)
            """
            lane_to_tr_item = []

            for tr_item in track.tr_track:  # type: TrTrackItem
                if tr_item.lr_item:
                    lr_scene_img = track.lr_track.best_image
                    lr_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
                    lane_set = self._lane_detector.get_lane2(lr_rect.fl, lr_rect.ft, lr_rect.fr, lr_rect.fb)
                    if len(lane_set) != 1:
                        lane_to_tr_item.append(('no_lane', tr_item))
                        continue

                    lane_name = lane_set.pop()
                    lane_to_tr_item.append((lane_name, tr_item))

            return lane_to_tr_item

        lane_to_tr_item_map = get_parts()

        for rule in self._rules:
            evidence = rule.check(track, base, lane_to_tr_item_map)
            if evidence:
                handler = functools.partial(self.handle_evidence, base, evidence)
                rule.add_video(handler)
                break

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> bool
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def collect_stats(self, json_handler):
        """
        :type json_handler: (dict) -> None
        :rtype: None
        """
        total_fails = self._stats.lr_absent + self._stats.lr_invalid + self._stats.wa_absent + self._stats.ra_cam_speed
        jnode = {
            'type': 'spot-speed-waga',
            'name': 'spot-speed-waga.{}'.format(self._identity.id),
            'stat': {
                'assert-track': {
                    'count': [self._stats.total, self._stats.total - total_fails, total_fails],
                    'wa-absent': self._stats.wa_absent,
                    'lr-absent': self._stats.lr_absent,
                    'lr-invalid': self._stats.lr_invalid,
                    'racam-speed-invalid': self._stats.ra_cam_speed
                },
                'emit-event': {
                    'count': [
                        self._stats.emit_total,
                        self._stats.emit_total - self._stats.emit_fail,
                        self._stats.emit_fail
                    ]
                }
            }
        }

        json_handler(jnode)
        map(lambda x: x.collect_stats(json_handler), self._rules)


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)
