import time
from os import rename
from ujson import dump as ujson_dump
from collections import defaultdict

from edge import eventlog
from edge_bbox.types import QuadI
from edge.com.bs import IBsCounters

from .types import StatServiceArgs
from edge.model.stats_provider import IStatsProvider
from ..model.models import Track


class IStatService:

    def __init__(self):
        pass

    @property
    def react(self):
        """
        :rtype: ReactCounters
        """
        raise NotImplementedError

    @property
    def event(self):
        """
        :rtype: EventCounters
        """
        raise NotImplementedError

    @property
    def bs(self):
        """
        :rtype: IBsCounters
        """
        raise NotImplementedError

    def add_providers(self, providers):
        """
        :type providers: list[IStatsProvider]
        """
        raise NotImplementedError


def _get_list_min_max_med(float_list):
    """
    :type float_list: list of float
    :rtype (float, float, float)
    """
    min_value = 0.0
    max_value = 0.0
    med_value = 0.0

    float_list_len = len(float_list)
    if float_list_len > 0:
        float_list.sort()
        min_value = float_list[0]
        max_value = float_list[-1]
        med_value = float_list[float_list_len / 2]

    return min_value, max_value, med_value


class BsCounters(IBsCounters):
    def __init__(self):
        IBsCounters.__init__(self)

        def _dict_factory():
            return {
                'blob-count': 0,
                'done-count': 0,
                'fail-count': 0
            }
        self._counters = defaultdict(_dict_factory)

    def inc_blob(self, chan):
        """
        :type chan: str
        """
        self._counters[chan]['blob-count'] += 1

    def inc_done(self, chan):
        """
        :type chan: str
        """
        self._counters[chan]['done-count'] += 1

    def inc_fail(self, chan):
        """
        :type chan: str
        """
        self._counters[chan]['fail-count'] += 1

    def serialize(self):
        return self._counters


class ReactCounters:
    def __init__(self):
        self.lrra_track_count = long(0)
        self.lr_track_count = long(0)
        self.ra_track_count = long(0)
        self.track_count = long(0)

        self.near_plate_w_list = []
        self.near_plate_h_list = []
        self.far_plate_w_list = []
        self.far_plate_h_list = []

    def reset(self):
        self.near_plate_w_list[:] = []
        self.near_plate_h_list[:] = []
        self.far_plate_w_list[:] = []
        self.far_plate_h_list[:] = []

    def inc_plate_quad(self, catch_quad, lost_quad):
        """
        :type catch_quad: QuadI
        :type lost_quad: QuadI
        """
        catch_rect_w, catch_rect_h = catch_quad.wh
        lost_rect_w, lost_rect_h = lost_quad.wh

        if catch_rect_w * catch_rect_h >= lost_rect_w * lost_rect_h:
            self.near_plate_w_list.append(catch_rect_w)
            self.near_plate_h_list.append(catch_rect_h)

            self.far_plate_w_list.append(lost_rect_w)
            self.far_plate_h_list.append(lost_rect_h)

        else:
            self.near_plate_w_list.append(lost_rect_w)
            self.near_plate_h_list.append(lost_rect_h)

            self.far_plate_w_list.append(catch_rect_w)
            self.far_plate_h_list.append(catch_rect_h)

    def inc_track(self, tr_track):
        """
        :type tr_track: Track
        """
        self.track_count += 1

        if bool(tr_track.lr_track) and bool(tr_track.ra_track):
            self.lrra_track_count += 1

        if bool(tr_track.lr_track):
            self.lr_track_count += 1

        if bool(tr_track.ra_track):
            self.ra_track_count += 1


class EventCounters:
    def __init__(self):
        self.fact_count = long(0)

    def inc_fact(self):
        self.fact_count += 1


class _StatService(IStatService):
    def __init__(self, twisted_reactor, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTime
        :type args: StatServiceArgs
        """
        IStatService.__init__(self)

        self._providers = list()

        self._react = ReactCounters()
        self._event = EventCounters()
        self._bs = BsCounters()

        self._stat_abspath = args.stat_filepath
        self._stat_tmp_abspath = args.stat_filepath + '.tmp'
        self._stat_tout_sec = args.stat_tout_sec

        # reset on start
        twisted_reactor.callLater(0, self._report_counters)

        if args.stat_tout_sec > 0:
            twisted_reactor.callLater(args.stat_tout_sec, self._log_counters, twisted_reactor)

    @property
    def react(self):
        """
        :rtype: ReactCounters
        """
        return self._react

    @property
    def event(self):
        """
        :rtype: EventCounters
        """
        return self._event

    @property
    def bs(self):
        """
        :rtype: IBsCounters
        """
        return self._bs

    def add_providers(self, providers):
        """
        :type providers: list[IStatsProvider]
        """
        self._providers.extend(providers)

    def _log_counters(self, twisted_reactor):
        self._report_counters()
        self._reset()

        twisted_reactor.callLater(self._stat_tout_sec, self._log_counters, twisted_reactor)

    def _reset(self):
        self.react.reset()

    def _report_counters(self):
        sys_utc_time = time.time()

        npw_min, npw_max, npw_med = _get_list_min_max_med(self.react.near_plate_w_list)
        nph_min, nph_max, nph_med = _get_list_min_max_med(self.react.near_plate_h_list)
        fpw_min, fpw_max, fpw_med = _get_list_min_max_med(self.react.far_plate_w_list)
        fph_min, fph_max, fph_med = _get_list_min_max_med(self.react.far_plate_h_list)

        units = [{
            'type': 'stats',
            'name': 'stats',
            'stat': {
                'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                'tout-sec': self._stat_tout_sec
            }
        }]

        map(lambda provider: provider.collect_stats(units.append), self._providers)
        stat = {
            "units": units,
            "stat": {
                "mtime": [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                "tout-sec": self._stat_tout_sec
            },
            "react": {
                "count": {
                    "lr": self.react.lr_track_count,
                    "ra": self.react.ra_track_count,
                    "lrra": self.react.lrra_track_count,
                    "total": self.react.track_count
                },
                "plate": {
                    "near-w": [npw_min, npw_med, npw_max],
                    "near-h": [nph_min, nph_med, nph_max],
                    "far-w": [fpw_min, fpw_med, fpw_max],
                    "far-h": [fph_min, fph_med, fph_max]
                }
            },
            "event": {
                "count": self.event.fact_count
            },
            "bs": self.bs.serialize()
        }

        # noinspection PyBroadException
        try:
            with open(self._stat_tmp_abspath, 'w') as json_file:
                # Use ensure_ascii=false to output UTF-8.
                ujson_dump(stat, json_file, ensure_ascii=False)

            rename(self._stat_tmp_abspath, self._stat_abspath)

        except Exception:
            eventlog.log_x()


def init_service(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorTime
    :type args: StatServiceArgs
    :rtype: IStatService
    """
    return _StatService(twisted_reactor, args)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: IStatService | _StatService
    """
    pass
