import os
import ujson
from twisted.internet import inotify
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.com import fs

from .stats import TrackReactCounters, SceneReactCounters
from .._args import SceneReactArgs, TrackReactArgs
from .._models import ImageRef, JsonImRef, LrTrack, LrScene, RaScene


_LR_SCENE_STAT_DONE = 1
_LR_SCENE_STAT_TOUT = 2
_LR_SCENE_STAT_FAIL = 3


class ISceneReactProducer:
    def __init__(self):
        pass

    def set_handler(self, handler):
        raise NotImplementedError


class ITrackReactProducer:
    def __init__(self):
        pass

    def set_handler(self, handler):
        raise NotImplementedError


# noinspection PyUnusedLocal
def stub_scene_react_handler(im_ref, lr_scene, ra_scene):
    """
    :type im_ref: ImageRef
    :type lr_scene: LrScene
    :type ra_scene: RaScene
    """
    print 'stub_scene_react_handler(im_ref, lr_scene, ra_scene)'


# noinspection PyUnusedLocal
def stub_track_react_handler(lr_track):
    """
    :type lr_track: LrTrack
    """
    print 'stub_track_react_handler(lr_track)'


def _listen_react_dir(react_dir, react_handler):
    fs.silent_rm_r(react_dir)
    fs.silent_mkdir_r(react_dir)

    notifier = inotify.INotify()
    notifier.startReading()
    notifier.watch(FilePath(react_dir),
                   mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE,
                   callbacks=[react_handler])


def _load_react_args_jnode(react_args_filepath):
    """
    :type react_args_filepath: str | None
    :rtype: dict | None
    """
    react_jnode = None

    if not react_args_filepath:
        return react_jnode

    # noinspection PyBroadException
    try:
        with open(react_args_filepath) as json_file:
            react_jnode = ujson.load(json_file)

    except IOError as x:
        eventlog.log_d("fail: read %s (%s)" % (react_args_filepath, x))

    except ValueError as x:
        eventlog.log_d('fail: deserialize %s (%s)' % (react_args_filepath, x))

    except Exception:
        eventlog.log_x()

    return react_jnode


class _ReactContext:
    def __init__(self, rm_delayed_call, ignore_mc_scene_path, ignore_mg_scene_path, ignore_ra_scene_path):
        """
        :type rm_delayed_call: twisted.internet.interfaces.IDelayedCall
        :type ignore_mc_scene_path: bool
        :type ignore_mg_scene_path: bool
        :type ignore_ra_scene_path: bool
        """

        self.rm_delayed_call = rm_delayed_call
        self._ignore_mc_scene_path = ignore_mc_scene_path
        self._ignore_mg_scene_path = ignore_mg_scene_path
        self._ignore_ra_scene_path = ignore_ra_scene_path

        self.mc_scene_path = None
        self.mg_scene_path = None
        self.lr_scene_path = None
        self.ra_scene_path = None

    def _check_all(self):
        """
        :rtype: bool
        """
        if not self._ignore_mc_scene_path and not self.mc_scene_path:
            return False

        if not self._ignore_mg_scene_path and not self.mg_scene_path:
            return False

        if not self._ignore_ra_scene_path and not self.ra_scene_path:
            return False

        return self.lr_scene_path

    @staticmethod
    def set_mc_scene_path_and_check_all(mc_context, mc_scene_abs_path):
        """
        :type mc_context: _ReactContext
        :type mc_scene_abs_path: str
        :rtype: bool
        """
        mc_context.mc_scene_path = mc_scene_abs_path
        return mc_context._check_all()

    @staticmethod
    def set_mg_scene_path_and_check_all(mg_context, mg_scene_abs_path):
        """
        :type mg_context: _ReactContext
        :type mg_scene_abs_path: str
        :rtype: bool
        """
        mg_context.mg_scene_path = mg_scene_abs_path
        return mg_context._check_all()

    @staticmethod
    def set_lr_scene_path_and_check_all(lr_context, lr_scene_abs_path):
        """
        :type lr_context: _ReactContext
        :type lr_scene_abs_path: str
        :rtype: bool
        """
        lr_context.lr_scene_path = lr_scene_abs_path
        return lr_context._check_all()

    @staticmethod
    def set_ra_scene_path_and_check_all(ra_context, ra_scene_abs_path):
        """
        :type ra_context: _ReactContext
        :type ra_scene_abs_path: str
        :rtype: bool
        """
        ra_context.ra_scene_path = ra_scene_abs_path
        return ra_context._check_all()


class SceneReactProducer(ISceneReactProducer):
    # noinspection PyUnusedLocal
    def __init__(self, twisted_reactor, scene_react_args, stat_counters):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTime
        :type scene_react_args: SceneReactArgs
        :type stat_counters: SceneReactCounters
        """

        ISceneReactProducer.__init__(self)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stat_counters

        self._mc_dir = scene_react_args.mc_dir
        self._mg_dir = scene_react_args.mg_dir
        self._lr_dir = scene_react_args.lr_dir
        self._ra_dir = scene_react_args.ra_dir
        self._wait_tout_sec = float(scene_react_args.tout_msec) / 1000.0

        self._react_ts_to_context_map = {}
        self._react_handler = None
        self._last_react_ts = 0
        self._inotify = inotify.INotify()
        self._inotify.startReading()

    def set_handler(self, handler):
        # Allow to set the handler only once.
        if handler and not self._react_handler:

            if self._mc_dir:
                self._reset_inotify_watch(self._mc_dir, self._handle_mc_addscene_react)

            if self._mg_dir:
                self._reset_inotify_watch(self._mg_dir, self._handle_mg_addscene_react)

            if self._ra_dir:
                self._reset_inotify_watch(self._ra_dir, self._handle_ra_addscene_react)

            self._reset_inotify_watch(self._lr_dir, self._handle_lr_addscene_react)

            self._react_handler = handler
            return True

        return False

    # noinspection PyUnusedLocal
    def _handle_mc_addscene_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._mc_dir, self._handle_mc_addscene_react)
            eventlog.log_w('react dir was changed to %s' % self._mc_dir)
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_mc_scene_path_and_check_all)
        except Exception:
            eventlog.log_x()

    # noinspection PyUnusedLocal
    def _handle_mg_addscene_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._mg_dir, self._handle_mg_addscene_react)
            eventlog.log_w('react dir was changed to %s' % self._mg_dir)
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_mg_scene_path_and_check_all)
        except Exception:
            eventlog.log_x()

    # noinspection PyUnusedLocal
    def _handle_lr_addscene_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._lr_dir, self._handle_lr_addscene_react)
            eventlog.log_w('react dir was changed to %s' % self._lr_dir)
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_lr_scene_path_and_check_all)
        except Exception as x:
            eventlog.log_x()

    # noinspection PyUnusedLocal
    def _handle_ra_addscene_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch, self._ra_dir, self._handle_ra_addscene_react)
            eventlog.log_w('react dir was changed to %s' % self._ra_dir)
            return

        # noinspection PyBroadException
        try:
            self._handle_xx_addscene_react(filepath, _ReactContext.set_ra_scene_path_and_check_all)
        except Exception as x:
            eventlog.log_x()

    def _handle_xx_addscene_react(self, xx_filepath, set_xx_scene_path_and_check_all_handler):
        if not (isinstance(xx_filepath, FilePath) and xx_filepath.isfile()):
            return

        react_ts = int(os.path.basename(xx_filepath.path))
        if react_ts < self._last_react_ts:
            fs.silent_rm_r(xx_filepath.path)
            return

        react_context = self._react_ts_to_context_map.get(react_ts, None)
        if react_context:
            all_available = set_xx_scene_path_and_check_all_handler(react_context, xx_filepath.path)
            if all_available:
                react_context.rm_delayed_call.cancel()
                self._handle_addscene_react(react_ts)
        else:
            rm_dalayed_call = self._twisted_reactor.callLater(
                self._wait_tout_sec, self._handle_addscene_react, react_ts)

            react_context = _ReactContext(rm_dalayed_call, self._mc_dir is None,
                                          self._mg_dir is None, self._ra_dir is None)
            set_xx_scene_path_and_check_all_handler(react_context, xx_filepath.path)
            self._react_ts_to_context_map[react_ts] = react_context

    def _handle_addscene_react(self, react_ts):
        """
        :type react_ts: int
        """
        react_context = self._react_ts_to_context_map.pop(react_ts, None)
        all_available = not react_context.rm_delayed_call.active()

        if react_ts > self._last_react_ts:
            # any of mg_scene, lr_scene and ra_scene are received during _tout
            self._last_react_ts = react_ts
        elif all_available:
            # all mg_scene, lr_scene and ra_scene are received during _tout, but a bit later then expected
            pass
        else:
            # it's too late to handle the add_scene react
            return

        #
        # deserialize scene image
        #

        im_ref = None
        # noinspection PyBroadException
        try:
            mc_scene_jnode = _load_react_args_jnode(react_context.mc_scene_path)
            if mc_scene_jnode:
                im_ref = JsonImRef.deserialize_mc_scene(mc_scene_jnode)
                fs.silent_rm_r(react_context.mg_scene_path)
        except Exception:
            eventlog.log_x()

        finally:
            fs.silent_rm_r(react_context.mc_scene_path)

        if not im_ref:
            # noinspection PyBroadException
            try:
                mg_scene_jnode = _load_react_args_jnode(react_context.mg_scene_path)
                if mg_scene_jnode:
                    im_ref = JsonImRef.deserialize_mg_scene(mg_scene_jnode)

            except Exception:
                eventlog.log_x()

            finally:
                fs.silent_rm_r(react_context.mg_scene_path)

        #
        # deserialize lr-scene
        #

        lr_scene = None
        # noinspection PyBroadException
        try:
            lr_scene_jnode = _load_react_args_jnode(react_context.lr_scene_path)
            if lr_scene_jnode:
                lr_scene = LrScene.deserialize(lr_scene_jnode)

                if not im_ref:
                    im_ref = JsonImRef(
                        lr_scene.im_ref.ts, lr_scene.im_ref.url, lr_scene.im_ref.codec,
                        lr_scene.im_ref.width, lr_scene.im_ref.height)

                if lr_scene.stat == _LR_SCENE_STAT_DONE:
                    self._stat_counters.inc_lr_scene_item_count(len(lr_scene.scene))
                else:
                    lr_scene = None

        except Exception:
            eventlog.log_x()

        finally:
            fs.silent_rm_r(react_context.lr_scene_path)

        #
        # deserialize ra-scene
        #

        ra_scene = None
        try:
            ra_scene_jnode = _load_react_args_jnode(react_context.ra_scene_path)
            if ra_scene_jnode:
                ra_scene = RaScene.deserialize(ra_scene_jnode)
                self._stat_counters.inc_ra_scene_item_count(len(ra_scene.scene))

        except ValueError:
            eventlog.log_x()

        finally:
            fs.silent_rm_r(react_context.ra_scene_path)

        #
        # handle scene-react
        #

        if im_ref or lr_scene or ra_scene:
            # noinspection PyBroadException
            try:
                self._stat_counters.inc_scene_count()
                self._react_handler(im_ref, lr_scene, ra_scene)
            except Exception:
                eventlog.log_x()

    def _reset_inotify_watch(self, react_dir, handler):
        try:
            self._inotify.ignore(FilePath(react_dir))
        except KeyError:
            pass

        fs.silent_mkdir_r(react_dir)

        try:
            self._inotify.watch(
                FilePath(react_dir),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[handler])
        except inotify.INotifyError as x:
            eventlog.log_d('fail: reset inotify watch (%s)' % x)


class TrackReactProducer(ITrackReactProducer):
    def __init__(self, twisted_reactor, track_react_args, stat_counters):
        """
        :type track_react_args: TrackReactArgs
        :type stat_counters: TrackReactCounters
        """
        ITrackReactProducer.__init__(self)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stat_counters

        self._react_dir = track_react_args.react_dir

        self._react_handler = None
        self._inotify = inotify.INotify()
        self._inotify.startReading()

    def set_handler(self, handler):
        self._react_handler = handler
        self._reset_inotify_watch()

    # noinspection PyUnusedLocal
    def _handle_track_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w('react dir was changed to %s' % self._react_dir)
            return

        lr_track = None
        try:
            lr_track_jnode = _load_react_args_jnode(filepath.path)
            if lr_track_jnode:
                lr_track = LrTrack.deserialize(lr_track_jnode)

        except ValueError:
            eventlog.log_x()
            return

        finally:
            fs.silent_rm_r(filepath.path)

        if lr_track is None:
            self._stat_counters.inc_track_fail_count()
            return

        # noinspection PyBroadException
        try:
            self._stat_counters.inc_track_count()
            self._react_handler(lr_track)
        except Exception:
            eventlog.log_x()

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._react_dir))
        except KeyError:
            # it's ok, if filepath is not watched
            pass

        fs.silent_mkdir_r(self._react_dir)

        try:
            self._inotify.watch(
                FilePath(self._react_dir),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_track_react])
        except inotify.INotifyError:
            eventlog.log_x()
